package com.tzld.piaoquan.ad.engine.commons.score.model;

import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdActionFeature;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import org.apache.commons.math3.distribution.BetaDistribution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

public class UnionThompsonSamplingModel extends Model {

    private static final Logger LOGGER = LoggerFactory.getLogger(UnionThompsonSamplingModel.class);

    private Map<Long, Map<Long, AdActionFeature>> featureCache = new HashMap<>();

    @Override
    public int getModelSize() {
        return featureCache.size();
    }

    @Override
    public boolean loadFromStream(InputStreamReader in) throws Exception {
        Map<Long, Map<Long, AdActionFeature>> initModel = new HashMap<>();
        try (BufferedReader reader = new BufferedReader(in)) {

            String line = null;
            int size = 0;
            while ((line = reader.readLine()) != null) {
                String[] items = line.split("\t");
                if (items.length < 5) {
                    continue;
                }
                this.putFeature(
                        initModel,
                        new BigDecimal(items[0].trim()).longValue(),
                        new BigDecimal(items[1].trim()).longValue(),
                        new BigDecimal(items[2].trim()).doubleValue(),
                        new BigDecimal(items[3].trim()).doubleValue(),
                        new BigDecimal(items[4].trim()).doubleValue()
                );
                size++;
            }
            this.featureCache = initModel;
            LOGGER.info("[UNION THOMPSON SAMPLING MODEL LOAD] mode load over, size {}", size);
        } catch (
                Exception e) {
            LOGGER.info("[UNION THOMPSON SAMPLING MODEL LOAD] mode load error ", e);
        } finally {
            in.close();
        }

        return true;
    }

    private void putFeature(Map<Long, Map<Long, AdActionFeature>> modelCache, Long mediaId, Long creativeId, Double view, Double click, Double conversion) {
        Map<Long, AdActionFeature> mediaCache = modelCache.computeIfAbsent(mediaId, i -> new HashMap<>());
        AdActionFeature adActionFeature = new AdActionFeature();
        adActionFeature.setAdView(view);
        adActionFeature.setAdClick(click);
        adActionFeature.setAdConversion(conversion);
        mediaCache.put(creativeId, adActionFeature);

    }

    public double ctrScore(AdRankItem item) {
        Map<Long, AdActionFeature> creativeMap = this.featureCache.getOrDefault(item.getMediaId(), new HashMap<>());
        AdActionFeature adActionFeature = creativeMap.getOrDefault(item.getAdId(), new AdActionFeature());

        int alpha_cvr = (int) adActionFeature.getAdClick() + 10;
        int beta_cvr = 100 + (int) adActionFeature.getAdView() - (int) adActionFeature.getAdClick();
        return this.betaSampler(alpha_cvr, beta_cvr);
    }

    public double cvrScore(AdRankItem item) {
        Map<Long, AdActionFeature> creativeMap = this.featureCache.getOrDefault(item.getMediaId(), new HashMap<>());
        AdActionFeature adActionFeature = creativeMap.getOrDefault(item.getAdId(), new AdActionFeature());

        int alpha_cvr = (int) adActionFeature.getAdConversion() + 10;
        int beta_cvr = 100 + (int) adActionFeature.getAdClick() - (int) adActionFeature.getAdConversion();
        return this.betaSampler(alpha_cvr, beta_cvr);
    }

    public double betaSampler(double alpha, double beta) {
        BetaDistribution betaSample = new BetaDistribution(alpha, beta);
        return betaSample.sample();

    }
}