package com.tzld.piaoquan.ad.engine.service.score.impl;

import com.alibaba.fastjson.JSONObject;
import com.tzld.piaoquan.ad.engine.service.score.container.AdCreativeFeatureContainer;
import com.tzld.piaoquan.ad.engine.service.score.container.PidLambdaContainer;
import com.tzld.piaoquan.ad.engine.service.score.container.PidLambdaForCpcContainer;
import com.tzld.piaoquan.ad.engine.service.score.container.PidLambdaV2Container;
import com.tzld.piaoquan.ad.engine.service.score.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.service.score.param.BidRankRecommendRequestParam;
import com.tzld.piaoquan.ad.engine.service.score.param.UnionRankRecommendRequestParam;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdItemFeature;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.UserAdFeature;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScorerUtils;
import com.tzld.piaoquan.ad.engine.service.remote.FeatureRemoteService;
import com.tzld.piaoquan.ad.engine.service.score.RankService;
import com.tzld.piaoquan.ad.engine.service.score.convert.RequestConvert;
import com.tzld.piaoquan.ad.engine.service.score.param.RankRecommendRequestParam;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class RankServiceImpl implements RankService {

    private final static Logger log = LoggerFactory.getLogger(RankServiceImpl.class);
    DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd");
    DateTimeFormatter timeFormatter = DateTimeFormatter.ofPattern("yyyyMMdd HH:mm:ss");
    @Autowired
    FeatureRemoteService featureRemoteService;
    @Autowired
    RankServiceThompsonImpl rankServiceThompson;

    @Autowired
    AdCreativeFeatureContainer adCreativeFeatureContainer;
    @Value("${ad.model.cpm.max:200}")
    Double cpmMax=200d;
    @Value("${ad.model.cpm.min:30}")
    Double cpmMin=30d;
    @Value("${ad.pid.cpc.exp:30}")
    private String cpcPidExpCode;

    public AdRankItem adItemRank(RankRecommendRequestParam request){
        ScoreParam param= RequestConvert.requestConvert(request);
        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }
        Map<Long,List<AdPlatformCreativeDTO>> groupMap=request
                .getAdIdList()
                .stream()
                .collect(Collectors.groupingBy(creativeDTO -> creativeDTO.getCreativeId()));
        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(new ArrayList<>(groupMap.keySet()));
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
        //避免recommend-feature出问题
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(Long adId:groupMap.keySet()){
                AdRankItem item=new AdRankItem();
                item.setAdId(adId);
                item.setItemFeature(new AdItemFeature());
                rankItems.add(item);
            }
        }
        boolean inCpcPidExp=false;
        if (request.getAdAbExpArr() != null && request.getAdAbExpArr().size() != 0) {
            for (Map<String, Object> map : request.getAdAbExpArr() ) {
                if (map.getOrDefault("abExpCode", "").equals(cpcPidExpCode)) {
                    inCpcPidExp = true;
                }
            }
        }
        double lambda=-1d;
        if(inCpcPidExp){
            for(AdRankItem item:rankItems){
                try {
                    AdPlatformCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                    item.setBid1(dto.getBid1());
                    item.setBid2(dto.getBid2());
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(1d);
                }catch (Exception e){
                    log.error("rankItems info error itemId={}",item.getAdId());
                    e.printStackTrace();
                }
            }
        }else {
            for(AdRankItem item:rankItems){
                try {
                    AdPlatformCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                    item.setBid1(dto.getBid1());
                    item.setBid2(dto.getBid2());
                    lambda= PidLambdaForCpcContainer.getPidLambda(item.getAdId());
                    if(lambda<0){
                        item.setCpa(dto.getCpa());
                        item.setPidLambda(1);
                    }else {
                        if(dto.getCpa()>1&&lambda<=1){
                            lambda=2d;
                        }
                        item.setCpa(lambda);
                        item.setPidLambda(1d);
                    }
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(1d);
                }catch (Exception e){
                    log.error("rankItems info error itemId={}",item.getAdId());
                    e.printStackTrace();
                }
            }
        }

        // 兜底方案
        List<AdRankItem> rankResult;
        if (inCpcPidExp) {
            rankResult = rank(param, userAdFeature, rankItems, ScorerUtils.CVR_ADJUSTING);
        } else {
            rankResult = rank(param, userAdFeature, rankItems, ScorerUtils.BASE_CONF);
        }

        if (!CollectionUtils.isEmpty(rankResult)) {
            JSONObject object=new JSONObject();
            object.put("mid",request.getMid());
            object.put("adid",rankResult.get(0).getAdId());
            object.put("type",rankResult.get(0).getScore_type());
            object.put("pctr",rankResult.get(0).getCtr());
            object.put("pcvr",rankResult.get(0).getCvr());
            object.put("score",rankResult.get(0).getScore());
            object.put("pidLambda",rankResult.get(0).getPidLambda());
            object.put("lrsamples",rankResult.get(0).getLrSampleString());
            object.put("dataTime",currentTime.format(timeFormatter));
            object.put("creativeId",rankResult.get(0).getAdId());
            log.info("svc=adItemRank {}", JSONObject.toJSONString(object));
            object.remove("lrsamples");
            if(inCpcPidExp){
                AdPlatformCreativeDTO dto=groupMap.get(rankResult.get(0).getAdId()).get(0);
                object.put("cpa",dto.getCpa()*dto.getBid1());
                object.put("oCpa",dto.getCpa());
                log.info("svc=cpc_pid obj={}", JSONObject.toJSONString(object));
            }else {
                log.info("svc=pid_log obj={}", JSONObject.toJSONString(object));
            }
            return rankResult.get(0);
        }else {
            //空返回值
            return new AdRankItem();
        }
    }



    @Override
    public AdPlatformCreativeDTO adBidRank(BidRankRecommendRequestParam request) {

        ScoreParam param= RequestConvert.requestConvert(request);

        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }
        Map<Long,List<AdPlatformCreativeDTO>> groupMap=request
                .getCreativeList()
                .stream()
                .collect(Collectors.groupingBy(creativeDTO -> creativeDTO.getCreativeId()));


        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(new ArrayList<>(groupMap.keySet()));
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
        double lambda=-1d;
        for(AdRankItem item:rankItems){
            try {
                AdPlatformCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                lambda=PidLambdaContainer.getPidLambda(item.getAdId());
                if(lambda<0){
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(0.6);
                }else {
                    if(dto.getCpa()>1&&lambda<=1){
                        lambda=2d;
                    }
                    item.setCpa(lambda);
                    item.setPidLambda(1d);
                }

            }catch (Exception e){
                log.error("rankItems info error itemId={}",item.getAdId());
                e.printStackTrace();
            }
        }
//        for(AdRankItem item:rankItems){
//            try {
////                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()+"").get(0);
//                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
//                item.setBid1(dto.getBid1());
//                item.setBid2(dto.getBid2());
//                lambda=PidLambdaContainer.getPidLambda(item.getAdId());
//                if(lambda<0){
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(dto.getCpa()*0.6);
//                }else {
//                    if(dto.getCpa()>1&&lambda<=1){
//                        lambda=2d;
//                    }
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(lambda);
//                }
//
//            }catch (Exception e){
//                log.error("rankItems info error itemId={}",item.getAdId());
//                e.printStackTrace();
//            }
//        }
        List<AdRankItem> rankResult;
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(AdPlatformCreativeDTO dto:request.getCreativeList()){
                AdRankItem item=new AdRankItem();
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setAdId(dto.getCreativeId());
                item.setItemFeature(new AdItemFeature());
                lambda=PidLambdaContainer.getPidLambda(item.getAdId());
                if(lambda<0){
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(0.6);
                }else {
                    if(dto.getCpa()>1&&lambda<=1){
                        lambda=2d;
                    }
                    item.setCpa(lambda);
                    item.setPidLambda(1d);
                }
//                if(lambda<0){
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(dto.getCpa()*0.6);
//                }else {
//                    if(dto.getCpa()>1&&lambda<=1){
//                        lambda=2d;
//                    }
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(lambda);
//                }
//                item.setCpa(dto.getCpa());
//                item.setPidLambda(PidLambdaContainer.getPidLambda(item.getAdId()));
                rankItems.add(item);
            }
            rankResult=rankServiceThompson.rank(param, userAdFeature, rankItems,null);
        }else {
            rankResult=rank(param, userAdFeature, rankItems,ScorerUtils.BASE_CONF);
        }

        AdRankItem topItem=rankResult.get(0);

        AdPlatformCreativeDTO result= AdPlatformCreativeDTO
                .builder()
                .build();
        BeanUtils.copyProperties(topItem,result);
        result.setCreativeId(topItem.getAdId());
        result.setPctr(topItem.getCtr());
        result.setPcvr(topItem.getCvr());
//        result.setCreativeCode(groupMap.get(topItem.getAdId()+"").get(0).getCreativeCode());
        result.setCreativeCode(groupMap.get(topItem.getAdId()).get(0).getCreativeCode());
        double realECpm=0d;
            //经验值 待定
//        realECpm=topItem.getEcpm1();
        realECpm=topItem.getEcpm2();

        if(realECpm>cpmMax/1000d){
            realECpm=cpmMax/1000d;
        }
        if(realECpm<cpmMin/1000d){
            realECpm=cpmMin/1000d;
        }
        result.setEcpm2(realECpm);
        AdPlatformCreativeDTO dto=groupMap.get(topItem.getAdId()).get(0);
        JSONObject object=new JSONObject();
        object.put("mid",request.getMid());
        object.put("adid",result.getCreativeId());
        object.put("creativeCode",result.getCreativeCode());
        object.put("type",topItem.getScore_type());
        object.put("pctr",topItem.getCtr());
        object.put("pcvr",topItem.getCvr());
        object.put("lrsamples",topItem.getLrSampleString());
        object.put("pidLambda",topItem.getPidLambda());

        //临时加入供pid v2使用
        object.put("realECpm",realECpm);
        object.put("creativeId",result.getCreativeId());
        //CPA还原
        object.put("cpa",dto.getCpa()*dto.getBid1());
        object.put("oCpa",dto.getCpa());
        object.put("dataTime",currentTime.format(timeFormatter));
        log.info("svc=adBidRank {}", JSONObject.toJSONString(object));
        object.remove("lrsamples");
        log.info("svc=pid_log obj={}", JSONObject.toJSONString(object));
        return result;
    }

    @Override
    public AdPlatformCreativeDTO adBidRankNewPid(BidRankRecommendRequestParam request) {

        ScoreParam param= RequestConvert.requestConvert(request);
        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        UserAdFeature userAdFeature=featureRemoteService.getUserAdFeature(request.getMid());
        if(userAdFeature==null){
            userAdFeature=new UserAdFeature();
        }
        Map<Long,List<AdPlatformCreativeDTO>> groupMap=request
                .getCreativeList()
                .stream()
                .collect(Collectors.groupingBy(creativeDTO -> creativeDTO.getCreativeId()));


        Map<Long, AdRankItem> cache=adCreativeFeatureContainer.getAll(new ArrayList<>(groupMap.keySet()));
        List<AdRankItem> rankItems=Collections.emptyList();
        if(!cache.isEmpty()){
            rankItems=new LinkedList<>(cache.values());
        }
        double lambda=-1d;
        for(AdRankItem item:rankItems){
            try {
//                AdPlatformBidCreativeDTO dto=groupMap.get(item.getAdId()+"").get(0);
                AdPlatformCreativeDTO dto=groupMap.get(item.getAdId()).get(0);
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                lambda=PidLambdaV2Container.getPidLambda(item.getAdId());
//                if(lambda<0){
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(0.6);
//                }else {
//                    if(dto.getCpa()>1&&lambda<=1){
//                        lambda=2d;
//                    }
//                    item.setCpa(lambda);
//                    item.setPidLambda(1d);
//                }
                if(lambda<0){
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(dto.getCpa()*0.6);
                }else {
                    if(dto.getCpa()>1&&lambda<=1){
                        lambda=2d;
                    }
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(lambda);
                }

            }catch (Exception e){
                log.error("rankItems info error itemId={}",item.getAdId());
                e.printStackTrace();
            }
        }
        List<AdRankItem> rankResult;
        if(rankItems==null|| rankItems.size()==0){
            rankItems=new LinkedList<>();
            for(AdPlatformCreativeDTO dto:request.getCreativeList()){
                AdRankItem item=new AdRankItem();
                item.setBid1(dto.getBid1());
                item.setBid2(dto.getBid2());
                item.setAdId(dto.getCreativeId());
                item.setItemFeature(new AdItemFeature());
                lambda=PidLambdaV2Container.getPidLambda(item.getAdId());
//                if(lambda<0){
//                    item.setCpa(dto.getCpa());
//                    item.setPidLambda(0.6);
//                }else {
//                    if(dto.getCpa()>1&&lambda<=1){
//                        lambda=2d;
//                    }
//                    item.setCpa(lambda);
//                    item.setPidLambda(1d);
//                }
                if(lambda<0){
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(dto.getCpa()*0.6);
                }else {
                    if(dto.getCpa()>1&&lambda<=1){
                        lambda=2d;
                    }
                    item.setCpa(dto.getCpa());
                    item.setPidLambda(lambda);
                }
//                item.setCpa(dto.getCpa());
//                item.setPidLambda(PidLambdaContainer.getPidLambda(item.getAdId()));
                rankItems.add(item);
            }
            rankResult=rankServiceThompson.rank(param, userAdFeature, rankItems,null);
        }else {
            rankResult=rank(param, userAdFeature, rankItems,ScorerUtils.BASE_CONF);
        }

        AdRankItem topItem=rankResult.get(0);

        AdPlatformCreativeDTO result= AdPlatformCreativeDTO
                .builder()
                .build();
        BeanUtils.copyProperties(topItem,result);
        result.setCreativeId(topItem.getAdId());
        result.setPctr(topItem.getCtr());
        result.setPcvr(topItem.getCvr());
        result.setCreativeCode(groupMap.get(topItem.getAdId()).get(0).getCreativeCode());
        double realECpm=0d;
        //经验值 待定
//        realECpm=topItem.getEcpm1();
        realECpm=topItem.getEcpm2();

        if(realECpm>cpmMax/1000d){
            realECpm=cpmMax/1000d;
        }
        if(realECpm<cpmMin/1000d){
            realECpm=cpmMin/1000d;
        }
        result.setEcpm2(realECpm);
        AdPlatformCreativeDTO dto=groupMap.get(topItem.getAdId()).get(0);
        JSONObject object=new JSONObject();
        object.put("mid",request.getMid());
        object.put("adid",result.getCreativeId());
        object.put("creativeCode",result.getCreativeCode());
        object.put("type",topItem.getScore_type());
        object.put("pctr",topItem.getCtr());
        object.put("pcvr",topItem.getCvr());
        object.put("lrsamples",topItem.getLrSampleString());
        object.put("pidLambda",topItem.getPidLambda());
        //临时加入供pid v2使用
        object.put("realECpm",realECpm);
        object.put("creativeId",result.getCreativeId());
        //CPA还原
        object.put("cpa",dto.getCpa()*dto.getBid1());
        object.put("oCpa",dto.getCpa());
        object.put("dataTime",currentTime.format(timeFormatter));
        log.info("svc=adBidRankNewPid {}", JSONObject.toJSONString(object));
        object.remove("lrsamples");
        log.info("svc=pid_v2_log obj={}", JSONObject.toJSONString(object));
        return result;
    }

    public List<AdRankItem> rank (ScoreParam param, UserAdFeature userAdFeature, List<AdRankItem> rankItems,String configFile){
        List<AdRankItem> rankResult = ScorerUtils
                .getScorerPipeline(configFile)
                .scoring(param, userAdFeature, rankItems);
        return rankResult;
    }

    @Override
    public AdRankItem unionAdItemRank(UnionRankRecommendRequestParam request) {

        ScoreParam param= RequestConvert.requestConvert(request);

        LocalDateTime currentTime = LocalDateTime.now();
        int currentHour = currentTime.getHour();
        int dayOfWeek = currentTime.getDayOfWeek().getValue();
        param.getRequestContext().setHour(currentHour+"");
        param.getRequestContext().setWeek(dayOfWeek+"");
        param.getRequestContext().setRegion(request.getRegion().replace("省",""));
        param.getRequestContext().setCity(request.getCity().replace("市",""));
        param.getRequestContext().setDay(currentTime.format(dateFormatter));

        List<AdRankItem> rankResult;
        List<AdRankItem> rankItems = new LinkedList<>();
        for (AdPlatformCreativeDTO dto : request.getCreativeList()) {
            AdRankItem item = new AdRankItem();
            item.setBid1(dto.getBid1());
            item.setBid2(dto.getBid2());
            item.setAdId(dto.getCreativeId());
            item.setItemFeature(new AdItemFeature());
            item.setMediaId(request.getMediaId());
            item.setCpa(dto.getCpa());
            rankItems.add(item);
        }
        rankResult = rank(param, null, rankItems, ScorerUtils.UNION_THOMPSON_CONF);

        AdRankItem topItem=rankResult.get(0);

        Optional<AdPlatformCreativeDTO> first = request.getCreativeList().stream()
                .filter(dto -> Objects.equals(dto.getCreativeId(), topItem.getAdId()))
                .findFirst();
        if (first.isPresent()) {
            AdPlatformCreativeDTO dto = first.get();
            JSONObject object = new JSONObject();
            object.put("mid", request.getMid());
            object.put("adid", dto.getCreativeId());
            object.put("creativeCode", dto.getCreativeCode());
            object.put("type", topItem.getScore_type());
            object.put("pctr", topItem.getCtr());
            object.put("pcvr", topItem.getCvr());
            object.put("pidLambda", topItem.getPidLambda());
            // 临时加入供pid v2使用
            object.put("creativeId", dto.getCreativeId());
            // CPA还原
            object.put("cpa", dto.getCpa() * dto.getBid1());
            object.put("oCpa", dto.getCpa());
            object.put("dataTime", currentTime.format(timeFormatter));
            log.info("svc=pid_log obj={}", JSONObject.toJSONString(object));
            object.put("lrsamples", topItem.getLrSampleString());
            log.info("svc=adBidRank {}", JSONObject.toJSONString(object));
        }

        return topItem;
    }

}
